@@ -33,7 +33,14 @@
     Updated by      Richard Gooch   10-JUN-2000: Added compatibility entries
   for ISDN BRI driver. Thanks to Xuan Baldauf <xuan--lkml@baldauf.org>.
 
-    Last updated by Chris Rankin    22-JAN-2001: Added entries for I2C devices.
+    Updated by      Chris Rankin    22-JAN-2001: Added entries for I2C devices.
+
+    Updated by      Richard Gooch   30-JUL-2001: Added compatibility entries
+  for Stallion serial driver. Updated compatibility support for joystick to
+  take account of move from "joystick/%d" to "input/js%d".
+
+    Last updated by Richard Gooch   7-AUG-2001: Created <write_old_sd_name> and
+  supported SCSI discs 16 to 127.
 
 
 */
@@ -58,6 +65,9 @@
 
 
 static char get_old_ide_name (unsigned int major, unsigned int minor);
+static char *write_old_sd_name (char *buffer,
+				unsigned int major, unsigned int minor,
+				char *part);
 
 
 struct translate_struct
@@ -88,6 +98,10 @@ static struct translate_struct translate_table[] =
     {"misc/",      NULL},
     {"isdn/",      NULL},
     {"i2c/",       "i2c-%s"},
+    {"staliomem/", "staliomem%s"},  /*  Stallion serial driver               */
+    {"ttyE/",      "ttyE%s"},       /*  Stallion serial driver               */
+    {"cue/",       "cue%s"},        /*  Stallion serial driver               */
+    {"input/js",   "js%s"},         /*  Joystick driver                      */
     {NULL,         NULL}
 };
 
@@ -123,7 +137,7 @@ const char *get_old_name (const char *devname, unsigned int namelen,
 	compat_name = buffer;
     }
     else if ( (strncmp (devname, "scsi/", 5) == 0) &&
-	      (strcmp (devname + namelen - 7,"generic") == 0) )
+	      (strcmp (devname + namelen - 7, "generic") == 0) )
     {
 	sprintf (buffer, "sg%u", minor);
 	compat_name = buffer;
@@ -155,15 +169,13 @@ const char *get_old_name (const char *devname, unsigned int namelen,
     else if ( (strncmp (devname, "scsi/", 5) == 0) &&
 	      (strcmp (devname + namelen - 4, "disc") == 0) )
     {
-	sprintf ( buffer, "sd%c", 'a' + (minor >> 4) );
-	compat_name = buffer;
+	compat_name = write_old_sd_name (buffer, major, minor, "");
     }
     else if ( (strncmp (devname, "scsi/", 5) == 0) &&
 	      (strncmp (ptr = (strrchr (devname, '/') +1), "part", 4)
 	       == 0) )
     {
-	sprintf (buffer, "sd%c%s", 'a' + (minor >> 4), ptr + 4);
-	compat_name = buffer;
+	compat_name = write_old_sd_name (buffer, major, minor, ptr + 4);
     }
     else if ( (strncmp (devname, "ide/", 4) == 0) &&
 	      (strcmp (devname + namelen - 4, "disc") == 0) )
@@ -204,20 +216,6 @@ const char *get_old_name (const char *devname, unsigned int namelen,
 		 pty1[index >> 4], pty2[index & 0x0f]);
 	compat_name = buffer;
     }
-    else if (strncmp (devname, "joysticks/", 10) == 0)
-    {
-	int index;
-	char *type = "";
-
-	if (devname[10] == 'a') index = atoi (devname + 18);
-	else
-	{
-	    index = atoi (devname + 17);
-	    type = "d";
-	}
-	sprintf (buffer, "%sjs%d", type, index);
-	compat_name = buffer;
-    }
     return (compat_name);
 }   /*  End Function get_old_name  */
 
@@ -269,3 +267,33 @@ static char get_old_ide_name (unsigned int major, unsigned int minor)
     if (minor > 63) ++letter;
     return (letter);
 }   /*  End Function get_old_ide_name  */
+
+static char *write_old_sd_name (char *buffer,
+				unsigned int major, unsigned int minor,
+				char *part)
+/*  [SUMMARY] Write the old SCSI disc name to a buffer.
+    <buffer> The buffer to write to.
+    <major> The major number for the device.
+    <minor> The minor number for the device.
+    <part> The partition string. Must be "" for a whole-disc entry.
+    [RETURNS] A pointer to the buffer on success, else NULL.
+*/
+{
+    unsigned int disc_index;
+
+    if (major == 8)
+    {
+	sprintf (buffer, "sd%c%s", 'a' + (minor >> 4), part);
+	return (buffer);
+    }
+    if ( (major > 64) && (major < 72) )
+    {
+	disc_index = ( (major - 64) << 4 ) + (minor >> 4);
+	if (disc_index < 26)
+	    sprintf (buffer, "sd%c%s", 'a' + disc_index, part);
+	else sprintf (buffer, "sd%c%c%s",
+		      'a' + (disc_index / 26) - 1, 'a' + disc_index, part);
+	return (buffer);
+    }
+    return (NULL);
+}   /*  End Function write_old_sd_name  */
