@@ -185,11 +185,17 @@
     Updated by      Richard Gooch   27-JUL-2001: Bug fixes when execvp(3) fails
   (child did not exit).
 
-    Last updated by Richard Gooch   30-JUL-2001: Applied and cleaned up patch
+    Updated by      Richard Gooch   30-JUL-2001: Applied and cleaned up patch
   by Chris Rankin (regular subexpression support). Fixed potential buffer
   overrun problems by using <snprintf>. Thanks to Sebastian Krahmer
   <krahmer@suse.de>. Added "-np" switch. Thanks to John Fremlin.
 
+    Updated by      Richard Gooch   31-JUL-2001: Added support for DELETE
+  event. Added debug trace to <action_modload>.
+
+    Last updated by Richard Gooch   8-AUG-2001: Added support for recursively
+  reading config directories.
+
 
 */
 #define _GNU_SOURCE
@@ -233,6 +239,10 @@
 /*  Update only after changing code to reflect new protocol  */
 #define DEVFSD_PROTOCOL_REVISION_DAEMON  5
 
+#ifndef DEVFSD_NOTIFY_DELETE  /*  Temporary hack  */
+#  define DEVFSD_NOTIFY_DELETE        7
+#endif
+
 /*  Compile-time check  */
 #if DEVFSD_PROTOCOL_REVISION_KERNEL != DEVFSD_PROTOCOL_REVISION_DAEMON
 #error protocol version mismatch. Update your kernel headers
@@ -338,6 +348,8 @@ flag syslog_is_open = FALSE;
 static void setup_initial_entries ();
 static void read_config (CONST char *location, flag optional,
 			 unsigned long *event_mask);
+static void read_config_file (CONST char *path, flag optional,
+			      unsigned long *event_mask);
 static void process_config_line (CONST char *line, unsigned long *event_mask);
 static int process_yp_line (int instatus, char *inkey, int inkeylen,
 			    char *inval, int invallen, char *indata);
@@ -379,7 +391,7 @@ static flag expand_expression(char *output, unsigned int outsize,
                               const char *devname,
                               const regmatch_t *ex, unsigned int numexp,
                               FILE *errfp);
-static void expand_regexp (char *output, unsigned int outsize,
+static void expand_regexp (char *output, size_t outsize,
 			   const char *input, const char *devname,
 			   const regmatch_t *ex, unsigned int numex,
 			   FILE *errfp);
@@ -598,9 +610,6 @@ static void read_config (CONST char *location, flag optional,
     [RETURNS] Nothing.
 */
 {
-    FILE *fp;
-    char buf[STRING_LENGTH];
-
     if (location[0] == '+')
     {
 	/*  It's a YP map  */
@@ -622,14 +631,59 @@ static void read_config (CONST char *location, flag optional,
 	SYSLOG (LOG_ERR, "error reading map: \"%s\"\n", location + 1);
 	exit (1);
     }
-    if ( ( fp = fopen (location, "r") ) == NULL )
+    read_config_file (location, optional, event_mask);
+}   /*  End Function read_config   */
+
+static void read_config_file (CONST char *path, flag optional,
+			      unsigned long *event_mask)
+/*  [SUMMARY] Read a configuration database.
+    <path> The path to read the database from. If this is a directory, all
+    entries in that directory will be read (except hidden entries).
+    <optional> If TRUE, the routine will silently ignore a missing config file.
+    <event_mask> The event mask is written here. This is not initialised.
+    [RETURNS] Nothing.
+*/
+{
+    struct stat statbuf;
+    FILE *fp;
+    char buf[STRING_LENGTH];
+
+    if (stat (path, &statbuf) != 0)
     {
 	if ( optional && (errno == ENOENT) ) return;
 	SYSLOG (LOG_ERR, "error opening file: \"%s\"\t%s\n",
-		location, ERRSTRING);
+		path, ERRSTRING);
 	exit (1);
     }
-    if (trace_level > 2) fprintf (stderr, "Reading file: \"%s\"\n", location);
+    if ( S_ISDIR (statbuf.st_mode) )
+    {
+	DIR *dp;
+	struct dirent *de;
+
+	if ( ( dp = opendir (path) ) == NULL )
+	{
+	    SYSLOG (LOG_ERR, "error opening directory: \"%s\"\t%s\n",
+		    path, ERRSTRING);
+	    exit (1);
+	}
+	while ( ( de = readdir (dp) ) != NULL )
+	{
+	    char fname[STRING_LENGTH];
+
+	    if (de->d_name[0] == '.') continue;
+	    snprintf (fname, STRING_LENGTH, "%s/%s", path, de->d_name[0]);
+	    read_config_file (fname, optional, event_mask);
+	}
+	closedir (dp);
+	return;
+    }
+    if ( ( fp = fopen (path, "r") ) == NULL )
+    {
+	if ( optional && (errno == ENOENT) ) return;
+	SYSLOG (LOG_ERR, "error opening file: \"%s\"\t%s\n", path, ERRSTRING);
+	exit (1);
+    }
+    if (trace_level > 2) fprintf (stderr, "Reading file: \"%s\"\n", path);
     while (fgets (buf, STRING_LENGTH, fp) != NULL)
     {
 	char *line;
@@ -643,8 +697,8 @@ static void read_config (CONST char *location, flag optional,
     }
     fclose (fp);
     if (syslog_is_open)
-	SYSLOG (LOG_INFO, "read config file: \"%s\"\n", location);
-}   /*  End Function read_config   */
+	SYSLOG (LOG_INFO, "read config file: \"%s\"\n", path);
+}   /*  End Function read_config_file   */
 
 static void process_config_line (CONST char *line, unsigned long *event_mask)
 /*  [SUMMARY] Process a line from a configuration file.
@@ -711,6 +765,8 @@ static void process_config_line (CONST char *line, unsigned long *event_mask)
 	new->action.when = DEVFSD_NOTIFY_CHANGE;
     else if (strcasecmp (when, "CREATE") == 0)
 	new->action.when = DEVFSD_NOTIFY_CREATE;
+    else if (strcasecmp (when, "DELETE") == 0)
+	new->action.when = DEVFSD_NOTIFY_DELETE;
     else
     {
 	SYSLOG (LOG_ERR, "bad WHEN in config line: \"%s\"\n", line);
@@ -1049,6 +1105,8 @@ static void action_modload (const struct devfsd_notify_struct *info,
     argv[4] = device;
     argv[5] = NULL;
     snprintf (device, sizeof (device), "/dev/%s", info->devname);
+    if (trace_level > 1)
+	fprintf (stderr, "Calling modprobe with name: \"%s\"\n", device);
     if (modprobe != NULL)
     {
 	(*modprobe) (5, argv);
@@ -1411,7 +1469,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
       case AC_MKNEWCOMPAT:
 	if ( make_symlink (dest_name, compat_name) )
 	    if (trace_level > 1)
-		SYSLOG (LOG_INFO, "made symlink: \"%s\" for dev: %u,%u\n",
+		fprintf (stderr, "made symlink: \"%s\" for dev: %u,%u\n",
 			compat_name, info->major, info->minor);
 	break;
       case AC_RMOLDCOMPAT:
@@ -1422,7 +1480,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
 		    function_name, compat_name, ERRSTRING);
 	}
 	else if (trace_level > 1)
-	    SYSLOG (LOG_INFO, "unlinked: \"%s\"\n", compat_name);
+	    fprintf (stderr, "unlinked: \"%s\"\n", compat_name);
 	break;
     }
 }   /*  End Function action_compat  */
@@ -1508,7 +1566,7 @@ static void do_debug (int fd)
 	    type = "unregistered";
 	    break;
 	  case DEVFSD_NOTIFY_ASYNC_OPEN:
-	    type = "asynchonously opened";
+	    type = "asynchronously opened";
 	    break;
 	  case DEVFSD_NOTIFY_CLOSE:
 	    type = "closed";
@@ -1522,13 +1580,16 @@ static void do_debug (int fd)
 	  case DEVFSD_NOTIFY_CREATE:
 	    type = "create";
 	    break;
+	  case DEVFSD_NOTIFY_DELETE:
+	    type = "delete";
+	    break;
 	  default:
 	    fprintf (stderr, "Unknown type: %u\n", info.type);
 	    exit (1);
 	    /*break;*/
 	}
 	fprintf (stderr, "Entry: \"%s\"(len=%d) %s mode: %o uid: %d gid: %d\n",
-		 info.devname, strlen (info.devname), type,
+		 info.devname, (int) strlen (info.devname), type,
 		 (int) info.mode, (int) info.uid, (int) info.gid);
     }
     fprintf (stderr, "Error reading control file\t%s\n", ERRSTRING);
@@ -1720,7 +1781,7 @@ static flag make_symlink (const char *oldpath, const char *newpath)
 	if (errno == EEXIST)
 	{
 	    if (trace_level > 1)
-		SYSLOG (LOG_INFO, "symlink: \"%s\" already exists\n", newpath);
+		fprintf (stderr, "symlink: \"%s\" already exists\n", newpath);
 	}
 	else
 	    SYSLOG (LOG_ERR, "%s: error creating symlink: \"%s\"\t%s\n",
